/*
 *  07/14/2002
 *
 *  HtmlSitemap.java  - A NetObjects Fusion HTML Sitemap
 *  Copyright (C) 2002 Kai Toedter
 *  kai@toedter.com
 *  www.toedter.com
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package com.toedter.components;

import java.lang.*;

import java.util.*;
import java.io.*;
import com.netobjects.nfc.api.*;

/**
 *  HtmlSitemapV12 is a component for NetObjects Fusion that generates a HTML
 *  based sitemap from the internal site structure. The version number V12 in
 *  the class name is to avoid conflicts with existing older versions of this
 *  component within NetObjects Fusion
 *
 *@author     Kai Toedter
 *@version    1.2
 */
public class HtmlSitemapV12 extends com.netobjects.nfc.api.ComponentApp {
	// Constants
	final static String designModePictureName = "htmlsitemap-1.2.gif";

	// Property names & choices
	final static String fontPropertyName = "Font";
	final static String fontPropertyValues = "Set(Default|Arial, Helvetica, sans-serif|" +
			"Times New Roman, Times, serif|" +
			"Courier New, Courier, mono|" +
			"Georgia, Times New Roman, Times, serif|" +
			"Verdana, Arial, Helvetica, sans-serif)";
	private String fontString = "0";
	private final static String[] fontArray = {"Default",
			"Arial, Helvetica, sans-serif",
			"Times New Roman, Times, serif",
			"Courier New, Courier, mono",
			"Georgia, Times New Roman, Times, serif",
			"Verdana, Arial, Helvetica, sans-serif"};

	final static String sizePropertyName = "Font Size";
	final static String sizePropertyValues = "Set(None|-7|-6|-5|-4|-3|-2|-1|+1|+2|+3|+4|+5|+6|+7)";
	private String sizeString = "0";
	private final static String[] sizeArray = {"None", "-7", "-6", "-5", "-4", "-3", "-2", "-1",
			"+1", "+2", "+3", "+4", "+5", "+6", "+7"};

	final static String publishDirPropertyName = "Publish Directory";
	private String publishDirectory = "htmlsitemap";

	final static String imageDirPropertyName = "Image Directory";
	private String fileSeparator = System.getProperty("file.separator", "/");
	private String imageDirectory = "3dShadow";

	final static String rootNodePropertyName = "Root Node";
	private String rootNode = "Default";
	private DNode rootDNode = null;

	final static String excludedNodesPropertyName = "Excluded Nodes";
	private String excludedNodes = "";
	private Vector excludedNodesVector = null;

	final static String nodeTextPropertyName = "Node Text";
	final static String nodeTextPropertyValues = "Set(Title|Banner|Navigation Button)";
	private String nodeTextString = "0";

	final static String nodeColorPropertyName = "Node Colors";
	final static String nodeColorPropertyValues = "Set(Tree Depth|Subtree|Single Color)";
	private String nodeColorString = "0";

	final static String linkTargetPropertyName = "Link Target";
	private String linkTargetString = "";

	// Member variables
	private DDrawPicture designModePicture;
	private String siteMapString;

	// Assets
	private DAssetManager assetManager;
	String[] nodeImageFileNames = null;
	int[] nodeImageAssetIDs = null;
	String[] leafNodeImageFileNames = null;
	int[] leafNodeImageAssetIDs = null;
	int spaceImageAssetID = 0;
	int con1ImageAssetID = 0;
	int con2ImageAssetID = 0;
	int con3ImageAssetID = 0;

	// For tree computations
	private int treeDepth;
	private int maxTreeDepth;
	private int colorCount;
	private int subtreeCount;
	private String[] nodeLocations;
	private String[] leafNodeLocations;
	private String nodeLocation;
	private String leafNodeLocation;
	private String spaceLocation;
	private String con1Location;
	private String con2Location;
	private String con3Location;
	private String[] connections;


	//--------------------------------------------------------------------
	//---------------------- Methods needed for the API ------------------
	//--------------------------------------------------------------------

	/**
	 *  Description of the Method
	 *
	 *@param  assetManager  asset manager
	 *@param  aCodeBase     codebase
	 *@return               "HTML Site Map 1.2"
	 */
	public String onInstall(DAssetManager assetManager, String aCodeBase) {
		this.assetManager = assetManager;
		return "HTML Sitemap 1.2";
	}


	/**
	 *  Description of the Method
	 *
	 *@param  aLayout  the layout
	 *@param  aRect    the rectangle
	 *@param  aDrop    drag and drop
	 */
	public void onDrop(DLayout aLayout, DRect aRect, int aDrop) {
		try {
			// if the NFX is not being dropped by the user,
			// just return
			if (aDrop != ActivateState.Drop) {
				return;
			}

			designModePicture = new DDrawPicture();
			designModePicture.SetPositionRect(aRect.getLeft(), aRect.getTop(),
					aRect.getRight(), aRect.getBottom());
			designModePicture.setStretch(PictureStretchMode.DRAW_TILED);
			designModePicture.setImageFile(getNFXCodeBase() +
					designModePictureName);
			designModePicture.setAltTag("HTML Sitemap 1.2");
			designModePicture.setName("HTML Sitemap 1.2");
			designModePicture.setPictureText("HTML Sitemap 1.2");

			/*
			 *  DImage imageSizer = new DImage();
			 *  imageSizer.setImageFile(getNFXCodeBase() +
			 *  designModePictureName);
			 *  DSize currentImageSize = imageSizer.getSize();
			 *  designModePicture.SetSize(currentImageSize.getWidth(), currentImageSize.getHeight());
			 *  designModePicture.SetStyle(DrawObjectStyle.NonResizable,DrawObjectStyle.NonResizable);
			 *  designModePicture.setStretch(PictureStretchMode.DRAW_STRETCHED);
			 */
			designModePicture.setUsePictureText(true);

			// designModePicture.setHTMLBefore("<!-- ");
			// designModePicture.setHTMLAfter(" -->");

			aLayout.AddObject(designModePicture);
		} catch (Exception e) {
			System.out.println("An exception occurred while dropping the component");
		}

		return;
	}


	/**
	 *  Called before files get uploaded/published
	 */
	public void onFinalize() {
		if (nodeImageAssetIDs != null) {
			int colorCount = nodeImageAssetIDs.length;
			for (int i = 0; i < colorCount; i++) {
				assetManager.RemoveAsset(nodeImageAssetIDs[i]);
				assetManager.RemoveAsset(leafNodeImageAssetIDs[i]);
			}
		}
		if (spaceImageAssetID != 0) {
			assetManager.RemoveAsset(spaceImageAssetID);
		}
		if (con1ImageAssetID != 0) {
			assetManager.RemoveAsset(con1ImageAssetID);
		}
		if (con2ImageAssetID != 0) {
			assetManager.RemoveAsset(con2ImageAssetID);
		}
		if (con3ImageAssetID != 0) {
			assetManager.RemoveAsset(con3ImageAssetID);
		}

		nodeImageFileNames = null;
		nodeImageAssetIDs = null;
		leafNodeImageFileNames = null;
		leafNodeImageAssetIDs = null;
		spaceImageAssetID = 0;
		con1ImageAssetID = 0;
		con2ImageAssetID = 0;
		con3ImageAssetID = 0;
	}


	/**
	 *  This will create the Name-Value pair that is displayed in the property
	 *  inspector
	 *
	 *@param  names  the property names
	 *@param  types  the property types
	 */
	public void onInspect(CStringArray names, CStringArray types) {
		names.Set(nodeTextPropertyName);
		types.Set(nodeTextPropertyValues);
		names.Set(nodeColorPropertyName);
		types.Set(nodeColorPropertyValues);
		names.Set(fontPropertyName);
		types.Set(fontPropertyValues);
		names.Set(sizePropertyName);
		types.Set(sizePropertyValues);
		names.Set(imageDirPropertyName);
		types.Set(typeFile);
		names.Set(publishDirPropertyName);
		types.Set(typeString);
		names.Set(rootNodePropertyName);
		types.Set(typeString);
		names.Set(excludedNodesPropertyName);
		types.Set(typeString);
		names.Set(linkTargetPropertyName);
		types.Set(typeString);
	}


	/**
	 *  This method is called by NetObjects Fusion to retrieve and set property
	 *  values shown in the property inspector
	 *
	 *@param  event       property event
	 *@param  value       value
	 *@param  get         1, if values should be retrieved
	 *@param  apropIndex  not used
	 *@param  inspector   not used
	 *@return             the property value
	 */
	public String PropertyListener(String event, String value,
			int get, int apropIndex, IDInspector inspector) {
		if (get == 1) {
			if (event.equals(fontPropertyName)) {
				return fontString;
			} else if (event.equals(sizePropertyName)) {
				return sizeString;
			} else if (event.equals(publishDirPropertyName)) {
				return publishDirectory;
			} else if (event.equals(imageDirPropertyName)) {
				return imageDirectory;
			} else if (event.equals(rootNodePropertyName)) {
				return rootNode;
			} else if (event.equals(excludedNodesPropertyName)) {
				return excludedNodes;
			} else if (event.equals(nodeTextPropertyName)) {
				return nodeTextString;
			} else if (event.equals(nodeColorPropertyName)) {
				return nodeColorString;
			} else if (event.equals(linkTargetPropertyName)) {
				return linkTargetString;
			}
		} else {
			if (event.equals(fontPropertyName)) {
				fontString = value;
			} else if (event.equals(nodeTextPropertyName)) {
				nodeTextString = value;
			} else if (event.equals(linkTargetPropertyName)) {
				linkTargetString = value;
			} else if (event.equals(nodeColorPropertyName)) {
				nodeColorString = value;
			} else if (event.equals(sizePropertyName)) {
				sizeString = value;
			} else if (event.equals(imageDirPropertyName)) {
				String oldImageDirectory = imageDirectory;
				imageDirectory = value;
				int position = imageDirectory.lastIndexOf(fileSeparator);
				if (position > 1) {
					imageDirectory = imageDirectory.substring(0, position);
				}
				String startString = getNFXCodeBase() + "styles";
				startString = startString.toLowerCase();
				String testDirectory = imageDirectory.toLowerCase();

				if (testDirectory.startsWith(startString)) {
					position = imageDirectory.lastIndexOf(fileSeparator);
					imageDirectory = imageDirectory.substring(position + 1, imageDirectory.length());
				} else {
					DMessageBox msg = new DMessageBox();
					msg.Warning("Please choose a file in a subdirectory of the \n\"" +
							getNFXCodeBase() + "styles\" directory, e.g. a file in the\n\"" +
							getNFXCodeBase() + "styles" + fileSeparator + "3dShadow\" directory.");
					imageDirectory = oldImageDirectory;
				}
			} else if (event.equals(publishDirPropertyName)) {
				publishDirectory = value;
			} else if (event.equals(rootNodePropertyName)) {
				rootNode = value;
			} else if (event.equals(excludedNodesPropertyName)) {
				excludedNodes = value;

				excludedNodesVector = new Vector();
				StringTokenizer tokenizer = new StringTokenizer(excludedNodes, ",");

				while (tokenizer.hasMoreTokens()) {
					String token = tokenizer.nextToken();
					excludedNodesVector.addElement(token.trim());
				}
			}
		}
		return "";
	}


	/**
	 *  This method is called by NetObjects Fusion whenever it needs to generate
	 *  the HTML for the component
	 *
	 *@param  assetManager    the asset manager
	 *@param  publishContext  the publish context
	 */
	public void onPublish(DAssetManager assetManager, int publishContext) {
		onFinalize();

		connections = new String[20];
		String oldImageDirectory = imageDirectory;
		imageDirectory = getNFXCodeBase() + "Styles" + fileSeparator + imageDirectory;

		String nodeImageFilePrefix = imageDirectory + fileSeparator + "node";
		String leafNodeImageFilePrefix = imageDirectory + fileSeparator + "leafNode";
		colorCount = 0;
		while ((new File(nodeImageFilePrefix +
				Integer.toString(colorCount + 1) + ".gif")).exists()) {
			colorCount += 1;
		}

		nodeImageFileNames = new String[colorCount];
		nodeImageAssetIDs = new int[colorCount];
		leafNodeImageFileNames = new String[colorCount];
		leafNodeImageAssetIDs = new int[colorCount];
		nodeLocations = new String[colorCount];
		leafNodeLocations = new String[colorCount];

		for (int i = 0; i < colorCount; i++) {
			nodeImageFileNames[i] = nodeImageFilePrefix + Integer.toString(i + 1) + ".gif";
			nodeImageAssetIDs[i] = assetManager.AddAsset(nodeImageFileNames[i],
					IAssetType.Image, publishDirectory);
			leafNodeImageFileNames[i] = leafNodeImageFilePrefix + Integer.toString(i + 1) + ".gif";
			leafNodeImageAssetIDs[i] = assetManager.AddAsset(leafNodeImageFileNames[i],
					IAssetType.Image, publishDirectory);
			nodeLocations[i] = assetManager.GetAssetRelativeLocation(nodeImageAssetIDs[i],
					publishContext,
					IFileNameStyle.DosName);
			leafNodeLocations[i] = assetManager.GetAssetRelativeLocation(leafNodeImageAssetIDs[i],
					publishContext,
					IFileNameStyle.DosName);
		}

		String spaceImageFileName = imageDirectory + fileSeparator + "space.gif";
		spaceImageAssetID =
				assetManager.AddAsset(spaceImageFileName, IAssetType.Image, publishDirectory);
		String con1ImageFileName = imageDirectory + fileSeparator + "con1.gif";
		con1ImageAssetID =
				assetManager.AddAsset(con1ImageFileName, IAssetType.Image, publishDirectory);
		String con2ImageFileName = imageDirectory + fileSeparator + "con2.gif";
		con2ImageAssetID =
				assetManager.AddAsset(con2ImageFileName, IAssetType.Image, publishDirectory);
		String con3ImageFileName = imageDirectory + fileSeparator + "con3.gif";
		con3ImageAssetID =
				assetManager.AddAsset(con3ImageFileName, IAssetType.Image, publishDirectory);

		// This is the string that will be used to comment out the
		// design mode picture HTML code
		String theHTMLString;

		DNode siteNode = new DNode();
		DNode homeNode = siteNode.getHomeNode();

		spaceLocation = assetManager.GetAssetRelativeLocation(spaceImageAssetID,
				publishContext,
				IFileNameStyle.DosName);
		con1Location = assetManager.GetAssetRelativeLocation(con1ImageAssetID,
				publishContext,
				IFileNameStyle.DosName);
		con2Location = assetManager.GetAssetRelativeLocation(con2ImageAssetID,
				publishContext,
				IFileNameStyle.DosName);
		con3Location = assetManager.GetAssetRelativeLocation(con3ImageAssetID,
				publishContext,
				IFileNameStyle.DosName);

		for (int i = 0; i < 10; i++) {
			connections[i] = spaceLocation;
		}

		// Compute tree depth
		treeDepth = 0;
		maxTreeDepth = 0;
		rootDNode = null;
		computeTreeDepth(homeNode, assetManager, publishContext);
		if (rootDNode != null) {
			homeNode = rootDNode;
			treeDepth = 0;
			maxTreeDepth = 0;
			computeTreeDepth(homeNode, assetManager, publishContext);
		}

		treeDepth = 0;

		// Initialize the string before beginning
		siteMapString = "<table border=\"0\" cellspacing=\"0\" cellpadding=\"0\">";

		// Make the call to retrieve the page hierarchy
		subtreeCount = 0;
		generateTree(homeNode, assetManager, publishContext);
		siteMapString += "</table>";

		theHTMLString = "<!-- ";

		// Set the correct HTML for the component
		designModePicture.setHTMLBefore(theHTMLString);
		designModePicture.setHTMLAfter("--><!-- HTML Sitemap 1.2 by Kai Toedter (http://www.toedter.com) -->" + siteMapString);

		imageDirectory = oldImageDirectory;
	}


	//-----------------------------------------------------------------------
	//--------------------------- Component Methods -------------------------
	//-----------------------------------------------------------------------

	/**
	 *  Description of the Method
	 *
	 *@param  node            the current node
	 *@param  assetManager    the asset manager
	 *@param  publishContext  the publish context
	 */
	private void generateTree(DNode node, DAssetManager assetManager, int publishContext) {
		if (node == null) {
			return;
		}

		String nodeName = node.getNodeName();
		String nodeTitle = node.getTitleName();
		String nodeText = nodeTitle;

		// if node URL is empty the node is not supposed to be published
		if (!isExcluded(node, assetManager, publishContext)) {

			if (nodeTextString.equals("0")) {
				nodeText = nodeTitle;
			} else if (nodeTextString.equals("1")) {
				nodeText = node.getBannerName();
			} else if (nodeTextString.equals("2")) {
				nodeText = node.getButtonName();
			}

			// Add the current node to the asset manager
			// The AssetID is needed to determine the HREF to the node
			// int theNodeAssetID = assetManager.AddAsset(node);

			int numberOfChildren = node.getNumberOfChildren();

			if ((treeDepth - 1) >= 0) {
				DNode testNode = node.getNextSiblingNode();
				if (testNode != null) {
					boolean lastExcluded = true;
					while (testNode != null) {
						if (!isExcluded(testNode, assetManager, publishContext)) {
							lastExcluded = false;
						}
						testNode = testNode.getNextSiblingNode();
					}
					if (lastExcluded) {
						connections[treeDepth - 1] = con2Location;
					} else {
						connections[treeDepth - 1] = con3Location;
					}
				} else {
					connections[treeDepth - 1] = con2Location;
				}
			}
			if ((treeDepth - 2) >= 0) {
				if (connections[treeDepth - 2] == con3Location) {
					connections[treeDepth - 2] = con1Location;
				} else if (connections[treeDepth - 2] == con2Location) {
					connections[treeDepth - 2] = spaceLocation;
				}
			}

			if (treeDepth == 1) {
				subtreeCount += 1;
			}

			int locationIndex = 0;
			if (nodeColorString.equals("0")) {
				locationIndex = treeDepth % colorCount;
			} else if (nodeColorString.equals("1")) {
				locationIndex = subtreeCount % colorCount;
			}
			String nodeString = nodeLocations[locationIndex];
			if (numberOfChildren == 0) {
				nodeString = leafNodeLocations[locationIndex];
			} else {
				boolean leaf = true;
				DNode childNode = node.getFirstChildNode();
				for (int i = 0; i < numberOfChildren; i++) {
					if (!isExcluded(childNode, assetManager, publishContext)) {
						leaf = false;
					}
				}
				if (leaf) {
					nodeString = leafNodeLocations[locationIndex];
				}
			}

			if (numberOfChildren == 1 && node.getFirstChildNode() != null &&
					node.getFirstChildNode().getTitleName().equalsIgnoreCase(excludedNodes)) {
				nodeString = leafNodeLocations[locationIndex];
			}

			siteMapString += "<tr>";
			for (int i = 0; i < treeDepth; i++) {
				siteMapString += ("<td><img src=\"" + connections[i] + "\"></td>");
			}

			if (((treeDepth - 1) >= 0) && connections[treeDepth - 1] == con2Location) {
				connections[treeDepth - 1] = spaceLocation;
			}

			String colSpanString = "";
			if ((maxTreeDepth - treeDepth + 1) > 1) {
				colSpanString = " colspan=" + (maxTreeDepth - treeDepth + 1);
			}

			String targetString = "";
			if (!linkTargetString.equals("")) {
				targetString = " target=\"" + linkTargetString + "\"";
			}

			String nodeURL = getNodeRelativeURL(node, assetManager, publishContext);

			// test for external references
			if (nodeURL.equals("")) {
				if (nodeName.startsWith("Sitemap:")) {
					nodeURL = nodeName.substring(8);
				}
			}

			siteMapString += ("<td><img src=\"" + nodeString + "\"></td>" +
					"<td width=\"600\"" + colSpanString + "><a href=\"" +
					nodeURL + "\"" + targetString + ">");

			if (!fontString.equals("0") || !sizeString.equals("0")) {
				String faceString = "";
				String sizeString2 = "";
				if (!fontString.equals("0")) {
					faceString = " face=\"" + fontArray[Integer.parseInt(fontString)] + "\"";
				}
				if (!sizeString.equals("0")) {
					sizeString2 = " size=\"" + sizeArray[Integer.parseInt(sizeString)] + "\"";
				}
				siteMapString += "<font " + faceString + sizeString2 + ">";
			}
			siteMapString += nodeText;
			if (!fontString.equals("0") || !sizeString.equals("0")) {
				siteMapString += "</font>";
			}
			siteMapString += "</a></td></tr>";

			if (numberOfChildren > 0) {
				// recurse to the node's child
				treeDepth += 1;
				generateTree(node.getFirstChildNode(), assetManager, publishContext);
			}
		}

		if (node.getNextSiblingNode() != null &&
				(rootNode.equalsIgnoreCase("default") ||
				rootDNode != node)) {
			// recurse to the node's sibling
			generateTree(node.getNextSiblingNode(), assetManager, publishContext);
		} else {
			// bottomed out, return up the call stack
			treeDepth -= 1;
			return;
		}
	}


	/**
	 *  Computes the tree depth of the site
	 *
	 *@param  node            current node
	 *@param  assetManager    Description of the Parameter
	 *@param  publishContext  Description of the Parameter
	 */
	private void computeTreeDepth(DNode node, DAssetManager assetManager, int publishContext) {
		String nodeName = node.getNodeName();
		if (!rootNode.equalsIgnoreCase("default") &&
				rootNode.equalsIgnoreCase(node.getNodeName())) {
			rootDNode = node;
		}
		if (!isExcluded(node, assetManager, publishContext)) {
			if (node.getNumberOfChildren() != 0) {
				//recurse to the node's child
				treeDepth += 1;
				if (treeDepth > maxTreeDepth) {
					maxTreeDepth = treeDepth;
				}
				computeTreeDepth(node.getFirstChildNode(), assetManager, publishContext);
			}
		}

		if (node.getNextSiblingNode() != null &&
				(rootNode.equalsIgnoreCase("default") ||
				rootDNode != node)) {
			//recurse to the node's sibling
			computeTreeDepth(node.getNextSiblingNode(), assetManager, publishContext);
		} else {
			//bottomed out, return up the call stack
			treeDepth -= 1;
			return;
		}
	}


	/**
	 *  Get the relative URL of a given DNode
	 *
	 *@param  node            the DNode whose URL we want
	 *@param  am              the asset manager
	 *@param  publishContext  the publish context
	 *@return                 String that URL in String representation
	 */
	private String getNodeRelativeURL(DNode node, DAssetManager am, int publishContext) {
		// Add the current node to the asset manager
		// The AssetID is needed to determine the HREF to the node
		int nodeAssetID = am.AddAsset(node);
		String path = am.GetAssetRelativeLocation(nodeAssetID, publishContext, IFileNameStyle.DosName);
		am.RemoveAsset(nodeAssetID);
		return path;
	}


	/**
	 *  Gets the excluded attribute of the HtmlSiteMap object.
	 *
	 *@param  node            the node to be tested
	 *@param  assetManager    the asset manager
	 *@param  publishContext  the publish context
	 *@return                 true, if the node is excluded from the sitemap
	 */
	private boolean isExcluded(DNode node, DAssetManager assetManager, int publishContext) {
		String nodeURL = getNodeRelativeURL(node, assetManager, publishContext);
		String nodeName = node.getNodeName();
		if (nodeURL.trim().equals("") && !nodeName.startsWith("Sitemap:")) {
			return true;
		}

		if (nodeName == null || excludedNodesVector == null) {
			return false;
		}

		int size = excludedNodesVector.size();

		for (int i = 0; i < size; i++) {
			String testString = (String) (excludedNodesVector.elementAt(i));
			if (nodeName.equalsIgnoreCase(testString)) {
				return true;
			}
		}
		return false;
	}
}

